import { Err, Ok, Result } from "./result.ts";
import { IO, Program } from "./program.ts";
import { Session } from "./session.ts";

export class Shell implements Program {
    private session!: Session;

    async run(io: IO, _args: string[]): Promise<number> {
        this.session = io.session;

        while (true) {
            const dir = this.session.cwdString();
            const prompt = "user" + "@term:" + dir + "$ ";
            io.term.print(prompt);

            let line = "";
            while (true) {
                const ch = await io.input.getChar();
                if (ch === "\n") {
                    io.term.print(ch);
                    break;
                }
                if (ch === "\b") {
                    io.term.print("\\b");
                    line = line.slice(0, line.length - 1);
                    continue;
                }
                io.term.print(ch);
                line += ch;
            }
            if (line === "") {
                continue;
            }
            const res = this.lexLine(line);
            if (!res.ok) {
                io.term.print(`error: ${res.error}`);
                continue;
            }
            const [cmd, ...args] = res.value;
            const resText = this.runCommand(cmd, ...args);
            console.log({ resText });
            if (resText !== "") {
                io.term.print(`${resText}\n`);
            }
        }
    }

    private runCommand(...args: string[]) {
        const session = this.session;

        switch (args[0]) {
            case "":
                return "";
            case "pwd":
                return session.cwd();
            case "cd": {
                if (args.length > 2) {
                    return "cd: too many arguments";
                }

                const res = session.cd(args[1]);
                if (!res.ok) {
                    return `cd: ${res.error}`;
                }

                return "";
            }
            case "mkdir": {
                if (args.length === 1) {
                    return "mkdir: missing operand";
                }

                for (const dir of args.slice(1)) {
                    const res = session.mkdir(dir);
                    if (!res.ok) {
                        return `mkdir: ${res.error}`;
                    }
                }

                return "";
            }
            case "ls": {
                if (args.length === 1) {
                    const res = session.listFiles();
                    if (!res.ok) {
                        return res.error;
                    }
                    return res.value;
                }
                return args.slice(1)
                    .map((arg) => {
                        const res = session.listFiles(arg);
                        if (!res.ok) {
                            return res.error;
                        }
                        return res.value;
                    }).join("\n");
            }
            case "touch": {
                if (args.length === 1) {
                    return "touch: missing file operand";
                }
                for (const fn of args.slice(1)) {
                    session.touch(fn);
                }
                return "";
            }
            case "cat": {
                if (args.length === 1) {
                    return "cat: missing file operand";
                }
                return args.slice(1).map((v) => {
                    const r = session.cat(v);
                    return r.ok ? r.value : r.error;
                }).reduce((acc, v) => acc + "\n" + v);
            }
            case "echo": {
                if (args.length === 1) {
                    return "\n";
                }
                return args[1];
            }
            default:
                return `${args[0]}: Command not found`;
        }
    }

    private lexLine(text: string): Result<string[], string> {
        const segs: string[] = [];
        let seg = "";
        let inString = false;
        let escaped = false;
        let terminator: string = "";
        for (const ch of text) {
            if (!escaped && ch === "\\") {
                escaped = true;
                continue;
            }
            if (inString) {
                if (!escaped && ch === terminator) {
                    inString = false;
                } else {
                    seg += ch;
                }
            } else if (/["']/.test(ch)) {
                terminator = ch;
                inString = true;
            } else if (ch === " ") {
                segs.push(seg);
                seg = "";
            } else {
                seg += ch;
            }
            escaped = false;
        }
        if (escaped || inString) {
            return Err("malformed");
        }
        if (seg !== "") {
            segs.push(seg);
        }
        return Ok(segs);
    }
}
