export type Point = { x: number; y: number };
export const Point = (x: number, y: number): Point => ({ x, y });

export interface Term {
    size(): Point;
    cursor(): Point;
    setCursor(point: Point): void;
    clear(): void;
    write(text: string): void;
    print(text: string): void;
}

export interface KeyListener {
    hasPressedKey(): boolean;
    nextPress(): string;
}

export class CanvasTerm implements Term, KeyListener {
    private canvas!: HTMLCanvasElement;
    private cx!: CanvasRenderingContext2D;

    private _size: Point = { x: 80, y: 24 };
    private _cursor: Point = { x: 0, y: 0 };

    private fontSize!: Point;

    private keysPressed: string[] = [];

    private buffer = new Array<string>(80 * 24).fill(" ");

    public constructor() {}

    public init() {
        this.canvas = document.querySelector<HTMLCanvasElement>("#term")!;
        this.cx = this.canvas.getContext("2d")!;

        const fontHeight = 28;

        this.cx.font = `${fontHeight}px monospace`;

        this.fontSize = Point(
            this.cx.measureText("a").width,
            fontHeight,
        );

        const w = this._size.x * this.fontSize.x;
        const h = this._size.y * this.fontSize.y;

        this.canvas.width = w;
        this.canvas.height = h;

        //this.canvas.style.width = w.toString();
        //this.canvas.style.height = h.toString();

        this.cx.fillStyle = "black";
        this.cx.fillRect(0, 0, w, h);

        document.body
            .addEventListener("keydown", (ev) => {
                switch (ev.key) {
                    case "Enter":
                        this.keysPressed.push("\n");
                        break;
                    case "Backspace":
                        this.keysPressed.push("\b");
                        break;
                    case "Shift":
                    case "Control":
                    case "Alt":
                    case "AltGraph":
                        break;
                    case "/":
                    case "'":
                        ev.preventDefault();
                        this.keysPressed.push(ev.key);
                        break;
                    default:
                        this.keysPressed.push(ev.key);
                        break;
                }
            });
    }

    size(): Point {
        return this._size;
    }

    cursor(): Point {
        return this._cursor;
    }

    setCursor(point: Point): void {
        this._cursor.x = point.x;
        this._cursor.y = point.y;
    }

    clear(): void {
        this.cx.fillStyle = "black";
        this.cx.fillRect(
            0,
            0,
            this._cursor.x * this.fontSize.x,
            this._cursor.y * this.fontSize.y,
        );
    }

    private writeChar(ch: string) {
        this.buffer[this._cursor.y * 80 + this._cursor.x] = ch;
    }

    private writeBuffer() {
        this.cx.fillStyle = "black";
        this.cx.fillRect(0, 0, this.canvas.width, this.canvas.height);

        this.cx.fillStyle = "white";
        this.cx.font = `28px monospace`;
        for (let y = 0; y < 24; ++y) {
            const line = this.buffer.slice(y * 80, y * 80 + 80).join("");

            this.cx.fillText(
                line,
                0,
                y * this.fontSize.y + this.fontSize.y,
            );
        }
        this.cx.fillRect(
            this._cursor.x * this.fontSize.x,
            this._cursor.y * this.fontSize.y,
            this.fontSize.x,
            this.fontSize.y,
        );
    }

    write(text: string): void {
        const line = text.split("\n")[0].slice(
            0,
            this._size.x - this._cursor.x,
        );
        for (const ch of line) {
            this.writeChar(ch);
            this._cursor.x += 1;
        }
        this.writeBuffer();
    }

    print(text: string): void {
        const size = this.size();
        for (const ch of text) {
            const cursor = this.cursor();
            if (cursor.x >= size.x || ch === "\n") {
                this.setCursor(Point(0, cursor.y + 1));
            } else {
                this.write(ch);
            }
        }
        this.writeBuffer();
    }

    hasPressedKey(): boolean {
        return this.keysPressed.length > 0;
    }

    nextPress(): string {
        const key = this.keysPressed[0];
        this.keysPressed = this.keysPressed.slice(1);
        return key;
    }
}
